#ifndef _CONTEXT_H_
#define _CONTEXT_H_ //make idempotent
/* Contextual Exceptions classes for throwing exceptions that know where they
   came from.
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


/**
 * \file context.h
 * \brief Context-aware exception classes Context, ContextStack, ContException
 */

#include <vector>
#include <string>
#include <memory>

///Objects that track and use the current context
namespace context
{
   using std::vector;
   using std::string;
   using std::auto_ptr;

   /**
    * \brief Represents a program context.
    * \note The constructors and destructors have significant side effects
    */
   class Context
   {
   private:
      /// A string description of the context
      std::string description;
      
      // prevent use
      Context();
      
      // prevent use
      Context(Context const &);

   public:
      Context(const string &argd);

      ~Context();

      const std::string &getDescription() const;

      static void print(const vector<string> &list, size_t start=0, 
	 size_t end=0);
   };

   /**
    * \brief This is the context stack, a Singleton
    */
   class ContextStack
   {
   private:
      /// The actual stack object
      std::vector<Context *> stack;

      /// The variable used for getInstanceOf
      static ContextStack *instance;

   public:
      /**
       * \brief Static method to get a ContextStack
       * \return The ContextStack
       */
      static ContextStack &getInstanceOf()
      {
	 if (instance==NULL) instance=new ContextStack;
	 if (instance==NULL) throw "Argh!";
	 return *instance;
      }

      void push(Context &cont);

      void pop();

      void pop(Context &cx);
      
      auto_ptr< vector<string> > makeSVector() const;

      static void print();
   };

   /**
    * \brief An Exception that remembers the context it was thrown from
    */
   class ContException
   {
   private:
      /// a vector representing the context it was constructed in and thrown from
      auto_ptr <vector <string> > thrContext;

      /// The message the exception represents
      string msg;

   public:
      ContException();

      ContException(const string &argmsg);
      
      void print();
   };
};
#endif // end idempotent ifdef
