/* dutree  -- Graphical display of disk useage
   Copyright (C) 1992, 1996, 2000 Free Software Foundation, Inc.
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <gtk--/main.h>
#include <gtk--/window.h>
#include <gtk--/button.h>
#include <gtk--/box.h>
#include <gtk--/drawingarea.h>
#include <gtk--/style.h>
#include <gtk--/label.h>
#include <gtk--/widget.h>
#include <sigc++-1.0/sigc++/slot.h>
#include <sigc++-1.0/sigc++/signal_system.h>

#include "errno.h"
#include <sys/stat.h>
#include "save-cwd.h"
#include <unistd.h>
#include "savedir.h"
#include "config.h"
#include <fcntl.h>

#include <sstream>
#include <vector>
#include <cmath>
#include <cstdlib>
#include "context.h"
#include "drawerdep.h"
#include "drawervisitordep.h"
#include "colorpickerdep.h"
#include "dutree.h"
#include "rotatedrawer.h"
using context::Context;
using context::ContException;

/**
 * \brief Writes an exception to an ostream
 * \note a wrapper on Exception::write so normal syntax can be used
 */
std::ostream &operator<<(std::ostream &output, Exception &err)
{
   err.write(output);
   return output;
}

void destroy_handler()
{
   Gtk::Main::quit();
}



/**
 * \brief Sets a Gdk_Color to the rectangle's color
 * \param curcolor The color to set to the rectangle's color
 */
void EntryRect::setColor(Gdk_Color &curcolor) const
{
   curcolor.set_red(my_EnTree->r);
   curcolor.set_green(my_EnTree->g);
   curcolor.set_blue(my_EnTree->b);
}




/**
 * \brief Constructor
 * \param x_size The width of the imagemap
 * \param y_size The height of the imagemap
 */
TreeView::TreeView(gint x_size, gint y_size):
   drawing(false), rt(150,150,0.5,0.5)
{
   Context cx1("TreeView::TreeView(gint, gint)");
   size(x_size, y_size);
   //treeGen.max_depth=3;
   set_events(GDK_ENTER_NOTIFY_MASK| GDK_LEAVE_NOTIFY_MASK|GDK_EXPOSURE_MASK | 
      GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | 
      GDK_BUTTON_PRESS_MASK);
   treeGen.opt_separate_dirs=false;
   treeGen.set_watcher(*this);
   srandom(1);
   my_tree.reset(new EnTree);
   cur_tree=NULL;
   over_rect=NULL;
   firsttime=true;
   oldsize=0;
   //dwr.reset(new DarkFirst(get_window()));
}

/**
 * \brief Gets the EnTree that the mouse is currently over, or the top EnTree
 * \return The top EnTree or NULL on failure;
 */
EnTree const &TreeView::get_EnTree() const
{
   if (mouse_in && over_rect!=NULL ) return over_rect->get_EnTree();
   else if (cur_tree==NULL)
   {
      Exception err;
      throw err;
   }
   return *cur_tree;
}

/**
 * \brief Return the EnTree at the top
 */
EnTree const &TreeView::get_top_EnTree() const
{
   if (my_tree.get()==NULL)
   {
      Exception err;
      throw err;
   }
   return *my_tree;
}

/**
 * \brief Set the current EnTree for display
 * \param newcur The new current entry
 */
void TreeView::set_EnTree(EnTree const &newcur)
{
   if (&newcur==cur_tree) return;

   cur_tree=&newcur;
   rectangles.clear();
   EnTree_rect(cur_tree, false, 0,0,width(),height());
   draw_rect();
   seek_over();
   me_modified();
}

/**
 * \brief Updates the display when the file sizes change significantly
 * \param moddir The DirTree that contains file size info
 */
void TreeView::modified(DirTree const &moddir)
{
   if (cur_tree==NULL || my_tree.get()==NULL) return;
   if (oldsize==0)
   {
      oldsize=my_tree->size;
      return;
   }
   if (my_tree->size*100/oldsize>115)
   {
      rectangles.clear();
      EnTree_rect(cur_tree, false, 0, 0, 300, 300);
      draw_rect();
      oldsize=my_tree->size;
   }
}

/**
 * \brief Displays the space use graph
 * \note Calculates the space use graph, the first time
 * \param event ignored (but required by gtkmm interface)
 * \return true (event handled)
 */
gint TreeView::expose_event_impl(GdkEventExpose *event)
{
   Context cx1("TreeView::expose_event_impl()");
   try 
   {
      static int oldwidth=-1, oldheight=-1;
      if (cpick.get()==NULL)
	 cpick.reset(new AgeColor);

      if (dwr.get()==NULL)
	 dwr.reset(new PrettyDrawer(get_window()));
      if (firsttime || oldheight!=event->area.height || 
	 oldwidth!=event->area.width)
      {
	 {
	    Context cx2("Calculating directories");
	    firsttime=false;

	    cur_tree=my_tree.get();
	    treeGen.calc_EnTree(".", my_tree.get());
	 }
	 {
	    Context cx3("Drawing rectangles");
	    if (cur_tree==NULL) return true;
	    rectangles.clear();
	    EnTree_rect(cur_tree, false, 0,0,width(),height());
	    me_modified();
	    oldwidth=width();
	    oldheight=height();
	 }
      }
      draw_rect();
   }
   catch (ContException &cex)
   {
      cex.print();
   }
   return true;
}

gint TreeView::delayed_calc()
{
   cur_tree=my_tree.get();
   treeGen.calc_EnTree(".", my_tree.get());
   return false;
}

/**
 * \brief Called when a rectangle is clicked
 * \param rect The EntryRect that was clicked
 * \param button The mouse button that clicked, enumerated as in GTK
 */
bool TreeView::clicked(EntryRect const *rect, int button)
{
   if (cur_tree==NULL) return true;
   if (rect==NULL) return false;
   //left click 
   if (button == 1 && rect!=NULL) 
      set_EnTree(rect->get_EnTree());

   //right click
   if (button == 3)
   {
      if (cur_tree->parent!=NULL) set_EnTree(*cur_tree->parent);
   }
   return true;
}



/**
 * \brief Calculates the position and dimensions of the rectangles, based on
 * the EnTree
 * \param curent The EnTree to base calculations on 
 * \param vertical If true, draw the boxes vertically.
 * \param x The x coordinate of the available space
 * \param y The y coordinate of the available space
 * \param width The width of the available space
 * \param height The height of the available space
 */
void TreeView::EnTree_rect(EnTree const *curent, bool vertical, gint x, gint y, 
   gint width, gint height)
{
   if(curent==NULL) return;
   if (firsttime)
   {

   cout << curent->name << endl;
   cout << "x: " << x << " y: "<< y << " width: " << width << " height: " <<\
      height << endl;
   }

/* calculate the total size of children */
   uintmax_t childsize=0;
   for (EnTree *curchild=curent->child; curchild!=NULL; 
      curchild=curchild->next)
   {
      childsize+=curchild->size;
   }

/* calculate the size of the entry without children */
   uintmax_t selfsize=curent->size-childsize;
   
/* prepare the adjustable dimension and coordinate */ 
   gint &dimension=(vertical)?width:height;
   gint &coordinate=(vertical)?x:y;
   gint olddimension=dimension;
   dimension=(selfsize==0)?0:dimension*selfsize/curent->size;
   EntryRect cur(x, y, width+1, height, *curent, *this);
   rectangles.push_back(cur);
/* draw the rectangle */   

   
   if (firsttime)
   {
   cout << "x: " << x << " y: "<< y << " width: " << width << " height: " <<\
      height << endl << endl;
   }

/* ajust the coordiate and dimension for drawing children */
   coordinate+=dimension;
   dimension=olddimension-dimension;

   /* prepare the adjustable (opposite) dimension and coordinate for the child  */
   gint &child_dimension=(vertical)?height:width;
   olddimension=child_dimension;
   gint &child_coordinate=(vertical)?y:x;

   float fcoord=child_coordinate;

   gint oldx, oldy, oldwidth=-1, oldheight=-1;
   for (EnTree *curchild=curent->child; curchild!=NULL; 
      curchild=curchild->next)
   {
      float fdim;
      fdim=(curchild->size==0)?0:1.0f*child_dimension*curchild->size/childsize;
      child_dimension=static_cast<int>(fdim);
      if (x!=oldx || y!=oldy || width!=oldwidth || height!=oldheight) 
         EnTree_rect(curchild, vertical, x, y, width, height);
      oldx=x; 
      oldy=y; 
      oldwidth=width; 
      oldheight=height;
      fcoord+=fdim;
      child_coordinate=static_cast<int>(fcoord);
      if (firsttime)
      {
         cout << fcoord << "," << fdim<<endl;
      }
      child_dimension=olddimension;
   }
}

/**
 * \brief Draws the rectangles to the screen from the rectangles vector
 */
void TreeView::draw_rect()
{

   if (drawing) 
   {
      ContException err("Drawing on top!");
      throw err;
   }
   drawing=true;
   over_rect=NULL;

   for (size_t i=0; i<rectangles.size(); ++i)
   {
      (*dwr)(rectangles[i], (*cpick)(rectangles[i]));
   }
   drawing=false;
}

/**
 * \brief Determines whether the TreeView is ready for use
 */
bool TreeView::isReady()
{
   return (cur_tree!=NULL);
}

void TreeView::test_tree()
{
   EnTree *child;
   my_tree.reset(new EnTree());
   my_tree->set(100, "Top");
   child=new EnTree();
   child->set(25, "Yark");
   my_tree->add_child(child);
   child=new EnTree();
   child->set(12, "Park");
   my_tree->add_child(child);
   EnTree *childchild=new EnTree();
   childchild->set(6, "Park");
   child->add_child(childchild);
   child=new EnTree();
   child->set(13, "Sark");
   my_tree->add_child(child); 
}

static void
ensure_space (String *s, unsigned int size)
{
  if (s->alloc < size)
    {
      s->text = static_cast<char *>(xrealloc (s->text, size + 1));
      s->alloc = size;
    }
}
/* Initialize string S1 to hold SIZE characters.  */

static void
str_init (String **s1, unsigned int size)
{
  String *s;

  s = (String *) xmalloc (sizeof (struct String));
  s->text = static_cast<char *>(xmalloc (size + 1));

  s->alloc = size;
  *s1 = s;
}

/* Restore the previous working directory or exit.
   If CWD is null, simply call `chdir ("..")'.  Otherwise,
   use CWD and free it.  CURR_DIR_NAME is the name of the current directory
   and is used solely in failure diagnostics.  */
static void str_concatc (String *s1, const char *cstr)
{
  unsigned l1 = s1->length;
  unsigned l2 = strlen (cstr);
  unsigned l = l1 + l2;

  ensure_space (s1, l);
  strcpy (s1->text + l1, cstr);
  s1->length = l;
}

/* Truncate the string S1 to have length LENGTH.  */

static void
str_trunc (String *s1, unsigned int length)
{
  if (s1->length > length)
    {
      s1->text[length] = 0;
      s1->length = length;
    }
}

static void
pop_dir (struct saved_cwd *cwd, const char *curr_dir_name)
{
  if (cwd)
    {
      if (restore_cwd (cwd, "..", curr_dir_name))
   exit (1);
      free_cwd (cwd);
    }
  else if (chdir ("..") < 0)
    {
    //  error (1, errno, _("cannot change to `..' from directory %s"),
     //   quote (curr_dir_name));
    }
}

/* dutree  -- Graphical display of disk useage
   Copyright (C) 1992, 1996, 2000 Free Software Foundation, Inc.
   Copyrigt  (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Written by David MacKenzie <djm@gnu.ai.mit.edu>.  */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#ifndef errno
extern int errno;
#endif

#include <sys/types.h>
#include "pathmax.h"

#if HAVE_GETCWD
char *getcwd ();
#else
char *getwd ();
# define getcwd(Buf, Max) getwd (Buf)
#endif

/* Return the current directory, newly allocated, arbitrarily long.
   Return NULL and set errno on error. */

char *
xgetcwd ()
{
  char *ret;
  unsigned path_max;
  char buf[1024];

  errno = 0;
  ret = getcwd (buf, sizeof (buf));
  if (ret != NULL)
    return xstrdup (buf);
  if (errno != ERANGE)
    return NULL;

  path_max = (unsigned) PATH_MAX;
  path_max += 2;     /* The getcwd docs say to do this. */

  for (;;)
    {
      char *cwd = (char *) xmalloc (path_max);

      errno = 0;
      ret = getcwd (cwd, path_max);
      if (ret != NULL)
   return ret;
      if (errno != ERANGE)
   {
     int save_errno = errno;
     free (cwd);
     errno = save_errno;
     return NULL;
   }

      free (cwd);

      path_max += path_max / 16;
      path_max += 32;
    }
}

void
free_cwd (struct saved_cwd *cwd)
{
  if (cwd->desc >= 0)
    close (cwd->desc);
  if (cwd->name)
    free (cwd->name);
}

void (*xalloc_fail_func) PARAMS ((void)) = 0;

void
xalloc_die (void)
{
  if (xalloc_fail_func)
    (*xalloc_fail_func) ();
  //error (xalloc_exit_failure, 0, "%s", _(xalloc_msg_memory_exhausted));
  /* The `noreturn' cannot be given to error, since it may return if
     its first argument is 0.  To help compilers understand the
     xalloc_die does terminate, call exit. */
  exit (EXIT_FAILURE);
}



/* Return a freshly allocated string containing the filenames
   in directory DIR, separated by '\0' characters;
   the end is marked by two '\0' characters in a row.
   NAME_SIZE is the number of bytes to initially allocate
   for the string; it will be enlarged as needed.
   Use NAME_SIZE == -1 if you do not know the size.
   Return NULL (setting errno) if DIR cannot be opened, read, or closed.  */

#ifndef NAME_SIZE_DEFAULT
# define NAME_SIZE_DEFAULT 512
#endif

char *
savedir (const char *dir, off_t name_size)
{
  DIR *dirp;
  struct dirent *dp;
  char *name_space;
  size_t allocated = name_size; /* Overflow is checked indirectly below.  */
  size_t used = 0;
  int save_errno;

  dirp = opendir (dir);
  if (dirp == NULL)
    return NULL;

  /* Use the default if the size is not known.  Be sure "allocated"
     is at least `1' so there's room for the final NUL byte.
     Do not simply test name_size <= 0, because the initialization
     of "allocated" might have overflowed.  */
  if (name_size < 0 || allocated == 0)
    allocated = NAME_SIZE_DEFAULT;

  name_space = static_cast<char *>(xmalloc (allocated));

  errno = 0;
  while ((dp = readdir (dirp)) != NULL)
    {
      /* Skip "", ".", and "..".  "" is returned by at least one buggy
         implementation: Solaris 2.4 readdir on NFS filesystems.  */
      char const *entry = dp->d_name;
      if (entry[entry[0] != '.' ? 0 : entry[1] != '.' ? 1 : 2] != '\0')
   {
     size_t entry_size = strlen (entry) + 1;
     if (used + entry_size < used)
       xalloc_die ();
     if (allocated <= used + entry_size)
       {
         do
      {
        if (2 * allocated < allocated)
          xalloc_die ();
        allocated *= 2;
      }
         while (allocated <= used + entry_size);

         name_space = static_cast<char *>(xrealloc (name_space, allocated));
       }
     memcpy (name_space + used, entry, entry_size);
     used += entry_size;
   }
    }
  name_space[used] = '\0';
  save_errno = errno;
  if (CLOSEDIR (dirp) != 0)
    save_errno = errno;
  if (save_errno != 0)
    {
      free (name_space);
      errno = save_errno;
      return NULL;
    }
  return name_space;
}

/* Change to recorded location, CWD, in directory hierarchy.
   If "saved working directory", NULL))
   */

int
restore_cwd (const struct saved_cwd *cwd, const char *dest, const char *from)
{
  int fail = 0;
  if (cwd->desc >= 0)
    {
      if (fchdir (cwd->desc))
   {
     /*error (0, errno, "cannot return to %s%s%s",
       (dest ? dest : "saved working directory"),
       (from ? " from " : ""),
       (from ? from : ""));*/
     fail = 1;
   }
    }
  else if (chdir (cwd->name) < 0)
    {
//      error (0, errno, "%s", cwd->name);
      fail = 1;
    }
  return fail;
}

/* Record the location of the current working directory in CWD so that
   the program may change to other directories and later use restore_cwd
   to return to the recorded location.  This function may allocate
   space using malloc (via xgetcwd) or leave a file descriptor open;
   use free_cwd to perform the necessary free or close.  Upon failure,
   no memory is allocated, any locally opened file descriptors are
   closed;  return non-zero -- in that case, free_cwd need not be
   called, but doing so is ok.  Otherwise, return zero.  */

int
save_cwd (struct saved_cwd *cwd)
{
  static int have_working_fchdir = 1;

  cwd->desc = -1;
  cwd->name = NULL;

  if (have_working_fchdir)
    {
#if HAVE_FCHDIR
      cwd->desc = open (".", O_RDONLY | O_DIRECTORY);
      if (cwd->desc < 0)
   {
     //error (0, errno, "cannot open current directory");
     return 1;
   }

# if __sun__ || sun
      /* On SunOS 4, fchdir returns EINVAL if accounting is enabled,
    so we have to fall back to chdir.  */
      if (fchdir (cwd->desc))
   {
     if (errno == EINVAL)
       {
         close (cwd->desc);
         cwd->desc = -1;
         have_working_fchdir = 0;
       }
     else
       {
         error (0, errno, "current directory");
         close (cwd->desc);
         cwd->desc = -1;
         return 1;
       }
   }
# endif /* __sun__ || sun */
#else
# define fchdir(x) (abort (), 0)
      have_working_fchdir = 0;
#endif
    }

  if (!have_working_fchdir)
    {
      cwd->name = xgetcwd ();
      if (cwd->name == NULL)
   {
//   error (0, errno, "cannot get current directory");
     return 1;
   }
    }
  return 0;
}
/* Bits set in the FLAGS argument to `fnmatch'.  */
#define  FNM_PATHNAME   (1 << 0) /* No wildcard can ever match `/'.  */
#define  FNM_NOESCAPE   (1 << 1) /* Backslashes don't quote special chars.  */
#define  FNM_PERIOD  (1 << 2) /* Leading `.' is matched only explicitly.  */

#if !defined (_POSIX_C_SOURCE) || _POSIX_C_SOURCE < 2 || defined (_GNU_SOURCE)
#define  FNM_FILE_NAME  FNM_PATHNAME /* Preferred GNU name.  */
#define  FNM_LEADING_DIR   (1 << 3) /* Ignore `/...' after a match.  */
#define  FNM_CASEFOLD   (1 << 4) /* Compare without regard to case.  */
#endif

/* Value returned by `fnmatch' if STRING does not match PATTERN.  */
#define  FNM_NOMATCH 1



/* Match STRING against the filename pattern PATTERN, returning zero if
   it matches, nonzero if not.  */
int
fnmatch (const char *pattern, const char *string, int flags)
{
  register const char *p = pattern, *n = string;
  register char c;

/* Note that this evaluates C many times.  */
#define FOLD(c) ((flags & FNM_CASEFOLD) && ISUPPER ((unsigned char) (c)) \
                 ? tolower ((unsigned char) (c)) \
                 : (c))

  while ((c = *p++) != '\0')
    {
      c = FOLD (c);

      switch (c)
   {
   case '?':
     if (*n == '\0')
       return FNM_NOMATCH;
     else if ((flags & FNM_FILE_NAME) && *n == '/')
       return FNM_NOMATCH;
     else if ((flags & FNM_PERIOD) && *n == '.' &&
         (n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
       return FNM_NOMATCH;
     break;

   case '\\':
     if (!(flags & FNM_NOESCAPE))
       {
         c = *p++;
         if (c == '\0')
      /* Trailing \ loses.  */
      return FNM_NOMATCH;
         c = FOLD (c);
       }
     if (FOLD (*n) != c)
       return FNM_NOMATCH;
     break;

   case '*':
     if ((flags & FNM_PERIOD) && *n == '.' &&
         (n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
       return FNM_NOMATCH;

     for (c = *p++; c == '?' || c == '*'; c = *p++)
       {
         if (c == '?')
      {
        /* A ? needs to match one character.  */
        if (*n == '\0' || (*n == '/' && (flags & FNM_FILE_NAME)))
          /* There isn't another character; no match.  */
          return FNM_NOMATCH;
        else
          /* One character of the string is consumed in matching
             this ? wildcard, so *??? won't match if there are
             less than three characters.  */
          ++n;
      }
       }

     if (c == '\0')
       {
         if ((flags & (FNM_FILE_NAME | FNM_LEADING_DIR)) == FNM_FILE_NAME)
      for (; *n != '\0'; n++)
        if (*n == '/')
          return FNM_NOMATCH;
         return 0;
       }

     {
       char c1 = (!(flags & FNM_NOESCAPE) && c == '\\') ? *p : c;
       c1 = FOLD (c1);
       for (--p; *n != '\0'; ++n)
         if ((c == '[' || FOLD (*n) == c1) &&
        fnmatch (p, n, flags & ~FNM_PERIOD) == 0)
      return 0;
         else if (*n == '/' && (flags & FNM_FILE_NAME))
      break;
       return FNM_NOMATCH;
     }

   case '[':
     {
       /* Nonzero if the sense of the character class is inverted.  */
       register int noot;

       if (*n == '\0')
         return FNM_NOMATCH;

       if ((flags & FNM_PERIOD) && *n == '.' &&
      (n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
         return FNM_NOMATCH;

       noot = (*p == '!' || *p == '^');
       if (noot)
         ++p;

       c = *p++;
       for (;;)
         {
      register char cstart = c, cend = c;

      if (!(flags & FNM_NOESCAPE) && c == '\\')
        {
          if (*p == '\0')
            return FNM_NOMATCH;
          cstart = cend = *p++;
        }

      cstart = cend = FOLD (cstart);

      if (c == '\0')
        /* [ (unterminated) loses.  */
        return FNM_NOMATCH;

      c = *p++;
      c = FOLD (c);

      if ((flags & FNM_FILE_NAME) && c == '/')
        /* [/] can never match.  */
        return FNM_NOMATCH;

      if (c == '-' && *p != ']')
        {
          cend = *p++;
          if (!(flags & FNM_NOESCAPE) && cend == '\\')
            cend = *p++;
          if (cend == '\0')
            return FNM_NOMATCH;
          cend = FOLD (cend);

          c = *p++;
        }

      if (FOLD (*n) >= cstart && FOLD (*n) <= cend)
        goto matched;

      if (c == ']')
        break;
         }
       if (!noot)
         return FNM_NOMATCH;
       break;

     matched:;
       /* Skip the rest of the [...] that already matched.  */
       while (c != ']')
         {
      if (c == '\0')
        /* [... (unterminated) loses.  */
        return FNM_NOMATCH;

      c = *p++;
      if (!(flags & FNM_NOESCAPE) && c == '\\')
        {
          if (*p == '\0')
            return FNM_NOMATCH;
          /* XXX 1003.2d11 is unclear if this is right.  */
          ++p;
        }
         }
       if (noot)
         return FNM_NOMATCH;
     }
     break;

   default:
     if (c != FOLD (*n))
       return FNM_NOMATCH;
   }

      ++n;
    }

  if (*n == '\0')
    return 0;

  if ((flags & FNM_LEADING_DIR) && *n == '/')
    /* The FNM_LEADING_DIR flag says that "foo*" matches "foobar/frobozz".  */
    return 0;

  return FNM_NOMATCH;

#undef FOLD
}


struct exclude_st
  {
    char const **exclude;
    int exclude_alloc;
    int exclude_count;
  };

struct exclude_st *
new_exclude (void)
{
  struct exclude_st *ex = (struct exclude_st *) xmalloc (sizeof (struct exclude_st));
  ex->exclude_count = 0;
  ex->exclude_alloc = 64;
  ex->exclude = (char const **) xmalloc (ex->exclude_alloc * sizeof (char *));
  return ex;
}

int
excluded_filename (struct exclude_st const *ex, char const *f, int options)
{
  char const * const *exclude = ex->exclude;
  int exclude_count = ex->exclude_count;
  int i;

  for (i = 0;  i < exclude_count;  i++)
    if (fnmatch (exclude[i], f, options) == 0)
      return 1;

  return 0;
}

/* Allocate space for the hash structures, and set the global
   variable `htab' to point to it.  The initial hash module is specified in
   MODULUS, and the number of entries are specified in ENTRY_TAB_SIZE.  (The
   hash structure will be rebuilt when ENTRY_TAB_SIZE entries have been
   inserted, and MODULUS and ENTRY_TAB_SIZE in the global `htab' will be
   doubled.)  */

void
DirTree::hash_init (unsigned int modulus, unsigned int entry_tab_size)
{
   struct htab *htab_r;

   htab_r = (struct htab *)
    xmalloc (sizeof (struct htab) + sizeof (struct entry *) * modulus);

   htab_r->entry_tab = (struct entry *)
    xmalloc (sizeof (struct entry) * entry_tab_size);

   htab_r->modulus = modulus;
   htab_r->entry_tab_size = entry_tab_size;
   htab = htab_r;

   hash_reset ();
}

void DirTree::hash_reset (void)
{
  int i;
  struct entry **p;

  htab->first_free_entry = 0;

  p = htab->hash;
  for (i = htab->modulus; i > 0; i--)
    *p++ = NULL;
}

/* Insert INO and DEV in the hash structure HTAB, if not
   already present.  Return zero if inserted and nonzero if it
   already existed.  */

int DirTree::hash_insert2 (struct htab *ht, ino_t ino, dev_t dev)
{
  struct entry **hp, *ep2, *ep;
  hp = &ht->hash[ino % ht->modulus];
  ep2 = *hp;

  /* Collision?  */

  if (ep2 != NULL)
    {
      ep = ep2;

      /* Search for an entry with the same data.  */

      do
   {
     if (ep->ino == ino && ep->dev == dev)
       return 1;     /* Found an entry with the same data.  */
     ep = ep->coll_link;
   }
      while (ep != NULL);

      /* Did not find it.  */

    }

  ep = *hp = &ht->entry_tab[ht->first_free_entry++];
  ep->ino = ino;
  ep->dev = dev;
  ep->coll_link = ep2;     /* `ep2' is NULL if no collision.  */

  return 0;
}


/* Insert an item (inode INO and device DEV) in the hash
   structure in the global variable `htab', if an entry with the same data
   was not found already.  Return zero if the item was inserted and nonzero
   if it wasn't.  */

int DirTree::hash_insert (ino_t ino, dev_t dev)
{
  struct htab *htab_r = htab; /* Initially a copy of the global `htab'.  */

  if (htab_r->first_free_entry >= htab_r->entry_tab_size)
    {
      int i;
      struct entry *ep;
      unsigned modulus;
      unsigned entry_tab_size;

      /* Increase the number of hash entries, and re-hash the data.
    The method of shrimping and increasing is made to compactify
    the heap.  If twice as much data would be allocated
    straightforwardly, we would never re-use a byte of memory.  */

      /* Let `htab' shrimp.  Keep only the header, not the pointer vector.  */

      htab_r = (struct htab *)
   xrealloc ((char *) htab_r, sizeof (struct htab));

      modulus = 2 * htab_r->modulus;
      entry_tab_size = 2 * htab_r->entry_tab_size;

      /* Increase the number of possible entries.  */

      htab_r->entry_tab = (struct entry *)
   xrealloc ((char *) htab_r->entry_tab,
        sizeof (struct entry) * entry_tab_size);

      /* Increase the size of htab again.  */

      htab_r = (struct htab *)
   xrealloc ((char *) htab_r,
       sizeof (struct htab) + sizeof (struct entry *) * modulus);

      htab_r->modulus = modulus;
      htab_r->entry_tab_size = entry_tab_size;
      htab = htab_r;

      i = htab_r->first_free_entry;

      /* Make the increased hash table empty.  The entries are still
    available in htab->entry_tab.  */

      hash_reset ();

      /* Go through the entries and install them in the pointer vector
    htab->hash.  The items are actually inserted in htab->entry_tab at
    the position where they already are.  The htab->coll_link need
    however be updated.  Could be made a little more efficient.  */

      for (ep = htab_r->entry_tab; i > 0; i--)
   {
     hash_insert2 (htab_r, ep->ino, ep->dev);
     ep++;
   }
    }

  return hash_insert2 (htab_r, ino, dev);
}



/**
 * \brief determine the size of file or directory ENT. 
 * \return the size of ENT in units of 512-byte blocks. 
 * \param top One for external calls, zero for recursive calls.
 * \param last_dev The device that the parent directory of ENT is on.
 * \param depth is the number of levels (in hierarchy) down from a command
 * line argument.  Don't print if DEPTH > max_depth.
 * \param target The EnTree node to store results in
 * \note An important invariant is that when this function returns, the current
 * working directory is the same as when it was called.  
 */

uintmax_t DirTree::count_entry (const char *ent, int top, dev_t last_dev,
   int depth, EnTree *target)
{
   uintmax_t size;
   struct stat stat_buf;
   EnTree *unused_child=NULL;

   if (((top && opt_dereference_arguments)
      ? stat (ent, &stat_buf)
      : (*xstat) (ent, &stat_buf)) < 0)
   {
      //error (0, errno, "%s", path->text);
      exit_status = 1;
      return 0;
   }

   if (!opt_count_all
      && stat_buf.st_nlink > 1
      && hash_insert (stat_buf.st_ino, stat_buf.st_dev))
   return 0;        /* Have counted this already.  */

   size = ST_NBLOCKS (stat_buf);
   tot_size += size;
   if (opt_all || S_ISDIR(stat_buf.st_mode))
   {
      target->set_attribs(stat_buf);
   }
   if (S_ISDIR (stat_buf.st_mode))
   {
      unsigned pathlen;
      dev_t dir_dev;
      char *name_space;
      char *namep;
      struct saved_cwd *cwd;
      struct saved_cwd cwd_buf;
      struct stat e_buf;

      dir_dev = stat_buf.st_dev;

      /* Return `0' here, not SIZE, since the SIZE bytes
      would reside in the new filesystem.  */
      if (opt_one_file_system && !top && last_dev != dir_dev)
      return 0;      /* Don't enter a new file system.  */

#ifndef S_ISLNK
# define S_ISLNK(s) 0
#endif
      /* If we're traversing more than one level, or if we're
      dereferencing symlinks and we're about to chdir through a
      symlink, remember the current directory so we can return to
      it later.  In other cases, chdir ("..") works fine.
      Treat `.' and `..' like multi-level paths, since `chdir ("..")'
      wont't restore the current working directory after a `chdir'
      to one of those.  */
      if (strchr (ent, '/')
	 || DOT_OR_DOTDOT (ent)
	 || (xstat == stat
	 && lstat (ent, &e_buf) == 0
	 && S_ISLNK (e_buf.st_mode)))
      {
	 if (save_cwd (&cwd_buf))
	    exit (1);
	 cwd = &cwd_buf;
      }
      else
      cwd = NULL;

      if (chdir (ent) < 0)
      {
	 //error (0, errno, _("cannot change to directory %s"),
	 //path->text);
	 if (cwd)
	 free_cwd (cwd);
	 exit_status = 1;
	 /* Do return SIZE, here, since even though we can't chdir into ENT,
	 we *can* count the blocks used by its directory entry.  */
	 return propogate_up(opt_separate_dirs ? 0 : size, target);
      }

      name_space = savedir (".", stat_buf.st_size);
      if (name_space == NULL)
      {
	 //error (0, errno, "%s", path->text);
	 pop_dir (cwd, path->text);
	 exit_status = 1;
	 /* Do count the SIZE bytes.  */
	 return propogate_up(opt_separate_dirs ? 0 : size, target);
      }

      /* Remember the current path.  */

      str_concatc (path, "/");
      pathlen = path->length;

      for (namep = name_space; *namep; namep += strlen (namep) + 1)
      {
	 if (!excluded_filename (exclude, namep, 0))
	 {
	    EnTree *child=unused_child;
	    if (child==NULL && target !=NULL)
	    {
	       child=new EnTree();
	       target->add_child(child);
	    }
	    str_concatc (path, namep);
	    size += count_entry (namep, 0, dir_dev, depth + 1, child);
	    if (child!=NULL && child->name==NULL) 
	    unused_child=child;
	    else unused_child=NULL;
	    str_trunc (path, pathlen);
	 }
      }

      if(unused_child!=NULL)
	 unused_child->remove();
      free (name_space);
      pop_dir (cwd, path->text);

      str_trunc (path, pathlen - 1); /* Remove the "/" we added.  */
      if (depth <= max_depth || top)
      {
	 if (target==NULL) ;
	 else target->add(size, path->length > 0 ? path->text : "/");
      }
      return propogate_up(opt_separate_dirs ? 0 : size, target);
   }
   else if ((opt_all && depth <= max_depth) || top)
   {
      /* FIXME: make this an option.  */
      int print_only_dir_size = 0;
      if (!print_only_dir_size)
      {
	 if (target==NULL) ;
	 else target->add(size, path->length > 0 ? path->text : "/");
      }
   }
   return propogate_up(size, target);
}

/**
 * \brief Increases the parent size and returns the appropriate amount
 *
 * If up_immediate is enabled, increments all parents by size
 * \note Notifies the watcher that the DirTree has been modifed
 * \param size The size of the increment
 * \param target The entry to update the size of the parents of
 * \return 0 if parents have been incremented, size if parents have not
 */
uintmax_t DirTree::propogate_up(uintmax_t size, EnTree *target)
{
   if (target!=NULL && up_immediate) 
   {
      target->incr_up(size);
      if (my_watch!=NULL) my_watch->modified(*this);
      return 0;
   }
   else return size;
}

DirTree::DirTree()
{
   init();
}

/**
 * \brief Initializes all values to safe or standard defaults
 */
void DirTree::init()
{
   max_depth=INT_MAX;
   exclude=new_exclude();
   path=NULL;
   str_init(&path, INITIAL_PATH_SIZE);
   xstat=lstat;
   htab=NULL;
   tot_size=0;
   hash_init(INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);
   my_watch=NULL;

   opt_count_all=false;
   opt_dereference_arguments=false;
   opt_one_file_system=false;
   opt_separate_dirs=false;
   opt_all =false;
   up_immediate=true;
}

/**
 * \brief Wrapper for calculating entry sizes
 */
void DirTree::calc_EnTree(const char *target, EnTree *my_tree)
{
   Context cx1("DirTree::calc_EnTree(const char *, EnTree *)");
   count_entry(target, 1, 0, 0, my_tree);
}

void DirTree::set_watcher(Watcher<DirTree> &my_watch)
{
   this->my_watch=&my_watch;
}

class RotateDrawerCmd: public Command
{
private:
   RotateDrawer rotate;
   TreeView &tv;
public:
   RotateDrawerCmd(TreeView &atv)
      :
   rotate(atv), tv(atv)
   {}
   void operator()()
   {
      rotate.visit(*tv.getDrawer());
      tv.draw_rect();
   }
};

class RotateColorCmd: public Command
{
private:
   TreeView &tv;

public:
   RotateColorCmd(TreeView &atv)
      :
   tv(atv)
   {}
   void operator()()
   {
      static int i=0;
      switch(i%3)
      {
	 case 0:
	    tv.set_cpick(new ColorPicker);
	 break;

	 case 1:
	    tv.set_cpick(new GidColor);
	 break;

	 case 2:
	    tv.set_cpick(new AgeColor);
	 break;
      };
      ++i;
      tv.draw_rect();
   }
};

/**
 * \brief Constructor prepares widgets
 */
MainWindow::MainWindow():
   Gtk::Window(GTK_WINDOW_TOPLEVEL),
   child_button("Child"),
   next_button("Next"),
   prev_button("Prev"),
   parent_button("Parent"),
   curdir("./\n\n\n"),
   tview(300, 300),
   rbutton("Style", new RotateDrawerCmd(tview)),
   colorbutton("Color Scheme", new RotateColorCmd(tview))
{
// Handle destroy and clicked signals
   destroy.connect(slot(destroy_handler));
   set_border_width(10);
   child_button.clicked.connect(slot(this, &MainWindow::nav_child));
   next_button.clicked.connect(slot(this, &MainWindow::nav_next));
   parent_button.clicked.connect(slot(this, &MainWindow::nav_parent));
   prev_button.clicked.connect(slot(this, &MainWindow::nav_prev));
   
// pack widgets into boxes 

   add(packer);
   packer.pack_start(buttons, false);
   buttons.pack_start(parent_button);
   buttons.pack_start(rbutton);
   buttons.pack_start(colorbutton);
   buttons.pack_start(hbox);
   packer.pack_start(curdir, false);
   packer.pack_start(tview);
   tview.set_watcher(*this);
   curdir.show();
   parent_button.show();
   rbutton.show();
   buttons.show();
   hbox.show();
   tview.show();
   packer.show();
   show_all();
   
}

int MainWindow::delete_event_impl(GdkEventAny *)
{
   return false;
}

/**
 * \brief Displays information about the current EnTree compared to the top
 * \param cur The current EnTree
 * \param top The top EnTree
 */
void MainWindow::setInfo(EnTree const &cur, EnTree const &top)
{
   std::ostringstream dirinfo;
   dirinfo << cur.name << endl << "Blocks " << cur.size; 
   dirinfo << endl << cur.size*100 /top.size  << "% of "<<top.name;
   if (cur.parent !=NULL) dirinfo << endl << cur.size * 100 /cur.parent->size  << "% of parent";
   else dirinfo << endl;
   curdir.set_text(dirinfo.str());
}

/**
 * \brief Updates the info when the TreeView is modified
 */
void MainWindow::modified(const ImageMap &tv)
{
   if (tview.isReady()) setInfo(tview.get_EnTree(), tview.get_top_EnTree());
}


int main (int argc, char *argv[])
{
   Context cx1("main()");
   try
   {
      // all GTK applications must have a gtk_main(). Control ends here
      // and waits for an event to occur (like a key press or mouse event).
      Gtk::Main kit(argc, argv);

      MainWindow dutree;
      {
	 Context cx2("Update thread");
	 kit.run();
      }
   }
   catch (Exception &err)
   {
      cout << err;
   }
   catch (ContException &cxe)
   {
      cxe.print();
   }
   return 0;
}


