/* en_tree -- Class for storage of directory entries 
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include "en_tree.h"
#include <string.h>

/**
 * \brief Find the last sibling of this EnTree
 * \return The last EnTree
 */
EnTree *EnTree::seek_end()
{
   EnTree *current;
   for (current=this; current->next!=NULL; current=current->next);
   return current;
}

/**
 * \brief Add a child to this node
 * \param child The child to add
 */
int EnTree::add_child(EnTree *child)
{ 
   if (this->child==NULL) this->child=child;
   else
   {
      EnTree *end;
      end=this->child->seek_end();
      end->next=child;
      child->prev=end;
   }
   child->parent=this;
   return 0;
}

/**
 * \brief Removes this EnTree from its tree without deleting it
 */
void EnTree::remove()
{
   // remove reference in the previous node
   if (prev!=NULL) prev->next=next;

   // remove reference in the next node
   if (next!=NULL) next->prev=prev;

   // remove reference in the parent node
   if (parent!=NULL && parent->child==this) 
   parent->child=next;
}

/**
 * \brief Set the name and size of an EnTree node 
 * \param n_blocks The size of the entry
 * \param name The name of the entry
 */
void EnTree::set(uintmax_t n_blocks, const char *name)
{
   size=n_blocks;
   if (this->name!=NULL) free (this->name);
   if (name==NULL) this->name=NULL;
   else this->name=strdup(name);
}

/**
 * \brief Set the name and increments the size of an EnTree node
 * \param n_blocks The size to add
 * \param name The name to set
 */
void EnTree::add(uintmax_t n_blocks, const char *name)
{
   this->size+=n_blocks;
   if (this->name!=NULL) free (this->name);
   if (name==NULL) this->name=NULL;
   else this->name=strdup(name);
}

/**
 * \brief increases the size of all ancestors of this block
 * \param n_blocks The size to increase by
 */
void EnTree::incr_up(uintmax_t n_blocks)
{
  for (EnTree *cur=parent; cur!=NULL; cur=cur->parent)
  {
	  cur->size+=n_blocks;
  }
}
