/* dutree  -- Graphical display of disk useage
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifndef _EN_TREE_H_
#define _EN_TREE_H_ //make idempotent

#include <inttypes.h>
#include <stdlib.h>
#include <sys/stat.h>
#define xmalloc malloc

/**
 * \brief Represents a directory entry
 */
class EnTree
{
public:
   ///The name of the directory entry
   char *name;
   ///The size of the directory entry (and all children)
   uintmax_t size;
   ///The parent of the directory entry
   EnTree *parent;
   ///The first child of the directory entry
   EnTree *child;
   ///The next sibling of this directory entry
   EnTree *next;
   ///The previous sibling of this direcotry entry
   EnTree *prev;
   int r, g, b;

   ///The user id of this entry
   uid_t uid;

   ///The group id of the entry
   gid_t gid;

   ///The modification time of the entry
   time_t mtime;

   ///The access time of the entry
   time_t atime;

   EnTree()
      :
   uid(0), gid(0), mtime(0), atime(0)
   {
      init();
   }

   
/**
 * \brief Initialize a node to safe defaults 
 */
   void init()
   {
      //using std::random;
      EnTree *target=this;
      if (target==NULL) return;
      target->size=0;
      target->next=NULL;
      target->prev=NULL;
      target->parent=NULL;
      target->child=NULL;
      target->name=NULL;
      target->r=random()%65535;
      target->g=random()%65535;
      target->b=random()%65535;
   }
   /**
    * \brief Destroys this node and all its descendants
    */
   ~EnTree()
   {
      if (this==NULL) return;

      EnTree *nextcur;
      for (EnTree *cur=child; cur!=NULL; cur=nextcur)
      {
	 nextcur=cur->next;
	 delete cur;
      }
      free(name);
   }

   /**
    * \brief Find the end of the list
    */
   EnTree *seek_end();
   
   /* Add a child to a node */
   int add_child(EnTree *child);

   void remove();

   void set(uintmax_t n_blocks, const char *name);

   void add(uintmax_t n_blocks, const char *name);

   void incr_up(uintmax_t n_blocks);
   void set_attribs(struct stat const &stat_buf)
   {
      uid=stat_buf.st_uid;
      gid=stat_buf.st_gid;
      mtime=stat_buf.st_mtime;
      atime=stat_buf.st_atime;
   }
   gid_t getGid() const
   {
      return gid;
   }
   time_t getMTime() const
   {
      return mtime;
   }
   time_t getATime() const
   {
      return atime;
   }
   uintmax_t my_size() const
   {
      uintmax_t s=size;
      for (EnTree const *cur=child; cur!=NULL; cur=cur->next)
      {
	 s-=cur->size;
      }
      return s;
   }
};

#endif //_EN_TREE_H_
