/* gtkmm Image Map
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "imgmapdep.h"
/**
 * \brief Constructor with dimensions
 * \param aleft The left position of the Rectangle
 * \param atop The top position of the Rectangle
 * \param awidth The width of the Rectangle
 * \param aheight The height of the Rectangle
 */
Rectangle::Rectangle(float aleft, float atop, float awidth, float aheight)
   :
left(aleft), top(atop), width(awidth), height(aheight)
{}

/**
 * \brief Determines whether the Rectangle contains the given point
 * \param x The x coordinate
 * \param y The y coordinate
 */
bool Rectangle::contains(float x, float y) const
{
   if (x>=left && y>=top)
   {
      if(x-left<=width && y-top<=height)
         return true;
   }
   return false;
}

/**
 * \brief Copy constructor
 */
EntryRect::EntryRect(EntryRect const &arg):
   Rectangle(arg), my_EnTree(arg.my_EnTree), my_watcher(arg.my_watcher)
{
}

bool EntryRect::hover()
{
   if (my_watcher!=NULL) return my_watcher->region_hover(*this);
   return false;
}


/**
 * \brief Finds the region that the mouse is currently over and sets the
 * appropriate member
 */
void ImageMap::seek_over()
{
   int x=mouse_x;
   int y=mouse_y;
   if (over_rect!=NULL && over_rect->contains(x, y)) return;
   for (size_t i=0; i<rectangles.size(); ++i)
   {
      if (rectangles[i].contains(x, y)) 
      {
         over_rect=&rectangles[i];
         me_modified();
         return;
      }
   }
   over_rect=NULL;
}

/**
 * \brief Updates the information about which region the pointer is over
 * \param event The event containing information about the xy coordinates of
 * the pointer
 * return true (event handled)
 */
gint ImageMap::motion_notify_event_impl(GdkEventMotion *event)
{
   int x, y;
   GdkModifierType state;
   Gdk_Window window(event->window);
   if (event->window);
   if (event->is_hint)
      window.get_pointer (x,y,state);
   else
   {
      x=(int)event->x;
      y=(int)event->y;
      state=(GdkModifierType) event->state;
   }
   mouse_x=x;
   mouse_y=y;
   EntryRect *old_over=over_rect;
   seek_over();
   //if (over_rect!=old_over) hover(over_rect);
   if (over_rect!=old_over && over_rect!=NULL) over_rect->hover();
   DrawingArea::motion_notify_event_impl(event);
   return true;
}

/**
 * \brief Updates the display when a mouse button is clicked
 * \param event The information about the click event
 * \
 */
gint ImageMap::button_press_event_impl(GdkEventButton *event)
{
   DrawingArea::button_press_event_impl(event);
   clicked(over_rect, event->button);
   return true;
}

/**
 * \brief Updates the display when the mouse enters the map area
 * \param event Unused (but required by gtkmm)
 */
gint ImageMap::enter_notify_event_impl(GdkEventCrossing *event)
{
   mouse_in=true;
   entered();
   DrawingArea::leave_notify_event_impl(event);
   return true;
}

/**
 * \brief Updates the display when the mouse leaves the map area
 * \param event Unused (but required by gtkmm)
 */
gint ImageMap::leave_notify_event_impl(GdkEventCrossing *event)
{
   if (event==NULL) return false;
   if (!DrawingArea::leave_notify_event_impl(event))
   mouse_in=false;
   me_modified();
   return false;
}

void ImageMap::hover(EntryRect const *rect)
{
   me_modified();
}

void ImageMap::entered()
{
   me_modified();
}

void ImageMap::left()
{
   me_modified();
}

/**
 * \brief Sets the watcher that will watch this TreeView
 * \param my_watch The watcher
 */
void ImageMap::set_watcher(Watcher<ImageMap> &my_watch)
{
   this->my_watch=&my_watch;
}


