/* gtkmm Image Map
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
/**
 * \brief Represents a Rectangle's position and dimensions
 */
#ifndef _IMGMAP_H_ 
#define _IMGMAP_H_ //make idempotent
#include <gtk--/main.h>
#include <gtk--/drawingarea.h>
#include <vector>

class Rectangle
{
   public:
   float left, top, width, height;
   Rectangle(float left, float top, float width, float height);
   bool contains(float x, float y) const;
};

class RecTransform
{
   private:
   float offx, offy;
   float scalex, scaley;
   public:
   float power;
   RecTransform::RecTransform(float aoffx, float aoffy, float ascalex,
      float ascaley)
      :
   offx(aoffx), offy(aoffy), scalex(ascalex), scaley(ascaley), power(1)
   {}

   Rectangle operator()(Rectangle &myrect)
   {
      myrect.left*=(1-power+scalex*power);
      myrect.left+=offx*power;
      myrect.top*=(1-power+scaley*power);
      myrect.top+=offy*power;
      myrect.width*=(1-power+scalex*power);
      myrect.height*=(1-power+scaley*power);
      return myrect;
   }
};

template <typename T>
class RegionWatcher
{
   public:
   virtual bool region_click(T &region)
   {
      return false;
   }
   virtual bool region_hover(T &region)
   {
      return false;
   }
   virtual bool region_default(T &region)
   {
      return false;
   }
};

/**
 * \brief Interface for regions of an imagemap
 */
class MapRegion
{
   public:
/**
 * \brief Determines whether the MapRegion contains the given point
 * \param x The x coordinate
 * \param y The y coordinate
 */
   virtual bool contains(float x, float y)const=0 ;
   virtual bool hover()
   {
      return false;
   }
};


/**
 * \brief A Rectangle that represents a region of a map and has a directory
 * entry
 */
class EntryRect: public Rectangle, public MapRegion
{
   private:
   EnTree const *my_EnTree;
   public:
   RegionWatcher<EntryRect> *my_watcher;
   EntryRect(float left, float top, float width, float height, 
      EnTree const &arg_EnTree, RegionWatcher<EntryRect> &arg_watcher)
      :
      Rectangle(left, top, width, height), my_EnTree(&arg_EnTree), 
      my_watcher(&arg_watcher)
   {
      ;
   }
   EntryRect(EntryRect const &arg);
   /**
    * \brief Returns the EnTree for this rectangle
    */
   EnTree const &get_EnTree() const
   {
      if (my_EnTree==NULL)
      {
	 Exception err;
	 throw err;
      }
      return *my_EnTree;
   }
   void setColor(Gdk_Color &curcolor) const;
   bool contains(float x, float y) const
   {
      return Rectangle::contains(x,y);
   }
   bool hover();

   static EntryRect seek_failed;
};

/**
 * \brief A collection of regions that react to clicks and mouseovers 
 */
class ImageMap: public Gtk::DrawingArea 
{
private:
   Watcher<ImageMap> *my_watch;
   gint mouse_x, mouse_y;

protected:
   EntryRect *over_rect;
   bool mouse_in;
   std::vector<EntryRect> rectangles;

public:
   void seek_over();
   void set_watcher(Watcher<ImageMap> &my_watch);
   void me_modified()
   {
      if (my_watch!=NULL) my_watch->modified(*this);
   }
   gint motion_notify_event_impl(GdkEventMotion *event);
   gint button_press_event_impl(GdkEventButton *event);
   gint enter_notify_event_impl(GdkEventCrossing *event);
   gint leave_notify_event_impl(GdkEventCrossing *event);
   virtual bool clicked(EntryRect const *, int button)=0;
   virtual void hover(EntryRect const *);
   virtual void left();
   virtual void entered();
};

#endif //_IMGMAP_H_
