/**
 * Copyright (C) 2003 Aaron Bentley, all rights reserved
 */

/**
 * \brief Base Visitor class, for use with any type.  
 *
 * Can be subclassed to call T.accept(), or as a Virtual Visitor
 */
template <typename T>
class Visitor
{
   protected:
   //Defines the type that is being operated on
   typedef T ArgBase;

   /**
    * \brief Performs an action on an object based on ArgBase
    * \obj The object to perform the action on
    */
   virtual void do_visit(ArgBase &obj)=0;

   /**
    * \brief Provided to allow Visitor to have the same interface as SV
    *
    * Calls do_visit unconditionally
    */
   virtual void try_visit(ArgBase & obj)
   {
      do_visit(obj);
   }

   public:
   /**
    * \brief Perform the action on the object
    * \param obj The object to perform the action on
    */
   virtual void visit(ArgBase &obj)
   {
      do_visit(obj);
   }

   /**
    * \brief Allows function object syntax
    */
   void operator() (ArgBase &obj)
   {
      visit(obj);
   }
};

/**
 * \brief We use a namespace because we're declaring a class with a
 * really short name (SV)
 */
namespace visitor 
{
   template <typename S, typename P>
   class SV: public P
   {
      protected:
      /// The parent of this SV
      typedef P Parent;

      /// The base class of the argument
      typedef typename Parent::ArgBase ArgBase;

      /// The subclass that the argument may or may not be
      typedef S ArgSub;

      /**
       * \brief If ob can be cast to ArgSub, call do_visit, otherwise try the
       * parent
       * \param ob The object to try visiting
       */
      void try_visit(ArgBase &ob)
      {
	 // We dynamically cast pointers because this is probably faster than
	 // casting references
	 ArgSub *s=dynamic_cast<ArgSub *>(&ob);
	 if (s==NULL) Parent::try_visit(ob);
	 else do_visit(*s);
      }

      /**
       * \brief Do the action on objects of type SubType
       * \param ob The object to act on
       */
      virtual void do_visit(ArgSub &ob)=0;

      public:
      /**
       * \brief Performs the VirtualVisitor's action on the given object 
       * \param ob The object to perform the VirtualVisitor's action on
       */
      void visit(ArgBase &ob)
      {
	 try_visit(ob);
      }
   };

   /**
    * \brief Define a name for the virtual visitor type
    *
    * Type names must be given in reverse inheritance order, so that no type
    * appears before its subtype.  (The subtype would be treated as another type
    * if this happens.
    * \note Since the programmer must maintain the type order, it is
    * recommended to use only one typedef per base type.  Changing a central
    * typedef will cause compile errors in any visitors that ignore new subtypes
    */
//   typedef SV <Baz, SV<Bar, Visitor<Foo> > > FooVisitor;
}


