/**
 * Copyright (C) 2003 Aaron Bentley, all rights reserved
 */

/**
 * \brief Base Visitor class, for use with any type.  
 *
 * Supplies implementations of visit and try_visit, adds a requirement to 
 * supply do_visit<T>() in subclasses.  Can be subclassed to call T.accept(),
 * or as a Virtual Visitor
 */
template <typename T>
class Visitor
{
   protected:
   //Defines the type that is being operated on
   typedef T ArgBase;

   /**
    * \brief Performs an action on an object based on ArgBase
    * \obj The object to perform the action on
    */
   virtual void do_visit(ArgBase &obj)=0;

   /**
    * \brief Provided to allow Visitor to have the same interface as SV
    *
    * Calls do_visit unconditionally
    */
   virtual void try_visit(ArgBase & obj)
   {
      do_visit(obj);
   }

   public:
   /**
    * \brief Perform the action on the object
    * \param obj The object to perform the action on
    */
   virtual void visit(ArgBase &obj)
   {
      do_visit(obj);
   }
};

/**
 * \brief Objects related to Visitor
 */
namespace visitor 
{
   /**
    * \brief Subclass of Visitor for use as an abstract base class
    *
    * Supplies an implementation of try_visit() and visit(), and adds the 
    * requirement to provide a do_visit<S>() in subclasses
    * \param S The subclass of ArgBase to perform operations on
    * \param P The parent class (either another SV or a Visitor)
    */
   template <typename S, typename P>
   class SV: public P
   {
      protected:
      /// The parent of this SV
      typedef P Parent;

      /// The subclass that the argument may or may not be
      typedef S ArgSub;

      /// The base class of ArgSub
      typedef typename Parent::ArgBase ArgBase;

      /**
       * \brief If ob can be cast to ArgSub, call do_visit, otherwise try this
       * SV's parent
       * \param ob The object to try visiting
       */
      void try_visit(ArgBase &ob)
      {
	 // We dynamically cast pointers because this is probably faster than
	 // casting references
	 ArgSub *s=dynamic_cast<ArgSub *>(&ob);
	 if (s==NULL) Parent::try_visit(ob);
	 else do_visit(*s);
      }

      /**
       * \brief Add the requirement to define do_visit(ArgSub &)
       * \param ob The object to act on
       */
      virtual void do_visit(ArgSub &ob)=0;

      public:
      /**
       * \brief Performs the VirtualVisitor's action on the given object 
       * \param ob The object to perform the VirtualVisitor's action on
       */
      void visit(ArgBase &ob)
      {
	 try_visit(ob);
      }
   };

   /**
    * \brief Helper to produce a virtual visitor based on a current type and
    * the remainder of the typelist
    * \param Mytype The type that the SV will work on
    * \param ListRemain The remainder of the list 
    */
   template <typename Mytype, typename ListRemain>
   class TlVisitor_helper
   {
   public:
      ///The type produced when the next type is not Nil
      typedef SV<Mytype, typename TlVisitor_helper<typename ListRemain::Head, 
	 typename ListRemain::Tail>::Ret> Ret;
   };


   /**
    * \brief Specialization of the helper for the situation where Mytype is
    * the last type in the list
    * \param Mytype The type to create a plain Visitor for
    */
   template <typename Mytype>
   class TlVisitor_helper<Mytype, meta::Nil>
   {
   public:
      ///The type produced when the next type is Nil
      typedef Visitor<Mytype> Ret;
   };


   /**
    * \brief Produces the type of a visitor that visits all items in the
    * TypeList, starting from the head.
    *
    * \param List A TypeList of types to visit.  The last Non-Nil type will be
    * considered the base class to visit
    *
    * \return The Visitor subclass, as the Ret typedef
    */
   template <typename List>
   class TlVisitor
   {
   public:
      /**
       * \brief The "return value"; a subclass of Visitor that visits all
       * specified types
       */
      typedef typename TlVisitor_helper<typename List::Head, typename List::Tail>::Ret Ret;
   };
}
